/*****************************************************
 * daub.h
 * Shengpu Liu
 *
 * this file is to forward transfer data to wavelet
 * domain through daubechies wavelet lifting scheme
 * wavlet tranform or inverse transfer wavelet coeff.
 * to time domain
 *****************************************************/

#ifndef _DAUB_H
#define _DAUB_H

#include <iostream>
#include <fstream>
#include <math.h>
using namespace std;

#define sqrt2 sqrt(2)
#define sqrt3 sqrt(3)

template <class data_type>
class Daub
{
public:
  static const int forward=1, inverse=2;

protected:
  /*
    Split the vec into even and odd elements,
    where the even elements are in the first half
    of the vector and the odd elements are in the
    second half.
  */
  void split(data_type *vec, int N)
  {
    int start = 1;
    int end = N - 1;

    while (start < end) {
      for (int i = start; i < end; i = i + 2) {
	data_type tmp = vec[i];
	vec[i] = vec[i+1];
	vec[i+1] = tmp;
      }
      start = start + 1;
      end = end - 1;
    }
  }
  
  /*
    Merge the odd elements from the second half of the N
    element region in the array with the even elements in
    the first half of the N element region.  The result
    will be the combination of the odd and even elements
    in a region of length N.
  */
  void merge(data_type *vec, int N)
  {
    int half = N >> 1;
    int start = half-1;
    int end = half;

    while (start > 0) {
      for (int i = start; i < end; i = i + 2)
	{
	  data_type tmp = vec[i];
	  vec[i] = vec[i+1];
	  vec[i+1] = tmp;
	}
      start = start - 1;
      end = end + 1;
    }
  }

  void normalize(data_type *S, int N, int direction)
  {
    int half = N >> 1;

    for (int n = 0; n < half; n++) {
      if (direction == forward) {
	S[n] = (data_type)round(((sqrt3-1.0)/sqrt2) * S[n]);
	S[n+half] = (data_type)round(((sqrt3+1.0)/sqrt2) * S[n+half]);
      }
      else if (direction == inverse) {
	S[n] = (data_type)round(((sqrt3+1.0)/sqrt2) * S[n]);
	S[n+half] = (data_type)round(((sqrt3-1.0)/sqrt2) * S[n+half]);
      }
      else {
	cerr<<"daubechies::normalize: bad direction value"<<endl;
	exit(0);
      }
    }
  } // normalize
  
  void updateOne(data_type *S, int N, int direction)
  {
    int half = N >> 1;
    
    for (int n = 0; n < half; n++) {
      data_type updateVal = (data_type)round(sqrt3 * S[half+n]);
      if (direction == forward)
	S[n] = S[n] + updateVal;
      else if (direction == inverse)
	S[n] = S[n] - updateVal;
      else {
	cerr<<"daubechies::updateOne: bad direction value"<<endl;
	exit(0);
      }
    }
  } // updateOne

  void update(data_type *S, int N, int direction )
  {
    int half = N >> 1;
    
    for (int n = 0; n < half-1; n++) {
      if (direction == forward)
	S[n] = S[n] - S[half+n+1];
      else if (direction == inverse)
	S[n] = S[n] + S[half+n+1];
      else {
	cerr<<"daubechies::update: bad direction value"<<endl;
	exit(0);
      }
    }
    
    if (direction == forward)
      S[half-1] = S[half-1] - S[half];
    else if (direction == inverse)
      S[half-1] = S[half-1] + S[half];
  } // update

  void predict(data_type *S, int N, int direction )
  {
    int half = N >> 1;

    if (direction == forward)
      S[half] = (data_type)round(S[half] - (sqrt3/4.0)*S[0]
	- (((sqrt3-2)/4.0)*S[half-1]));
    else if (direction == inverse)
      S[half] = (data_type)round(S[half] + (sqrt3/4.0)*S[0] +
	(((sqrt3-2)/4.0)*S[half-1]));
    else
      cerr<<"daubechies::predict: bad direction value"<<endl;

    // predict, forward
    for (int n = 1; n < half; n++)
      if (direction == forward)
	S[half+n] = (data_type)round(S[half+n] - (sqrt3/4.0)*S[n] -
	  (((sqrt3-2)/4.0)*S[n-1]));
      else if (direction == inverse)
	S[half+n] = (data_type)round(S[half+n] + (sqrt3/4.0)*S[n] +
	  (((sqrt3-2)/4.0)*S[n-1]));
      else exit(0);
  }  // predict

  data_type round(double val)
  {
    data_type ck = (data_type)val;
    if (ck == val) return ck;

    if(val >= 0)
      return (data_type)(val + 0.5);
    return (data_type)(val - 0.5);
  }
  
public:
  void forwardTrans(data_type *vec, int N, int nlevel)
  {
    //for (int n = N; n > 1; n = n >> 1) {
    for(int i=0; i<nlevel; i++){
      int n = N>>i; //only trans nlevel steps
      split( vec, n );
      updateOne( vec, n, forward );  // update 1
      predict( vec, n, forward );
      update( vec, n, forward );     // update 2
      normalize( vec, n, forward );
    }
  } // forwardTrans
  
  /*
    Default two step Lifting Scheme inverse wavelet transform
    inverseTrans is passed the result of an ordered wavelet
    transform, consisting of an average and a set of wavelet
    coefficients.  The inverse transform is calculated
    in-place and the result is returned in the argument array.
  */
  void inverseTrans(data_type *vec, int N, int nlevel)
  {
    //for (int n = 2; n <= N; n = n << 1) {
    for(int i=1; i<=nlevel; i++){
      int n = N>>(nlevel-i); //only invtrans nlevel steps
      normalize( vec, n, inverse );
      update( vec, n, inverse );
      predict( vec, n, inverse );
      updateOne(vec, n, inverse );
      merge( vec, n );
    }
  } // inverseTrans  
};

#endif

