/**************************************************************
* lswt.c
* This file implements the lifting scheme wavelet transform 
* algorithm for wireless sensor network based on SOS
* It is designed as a component that gets input from a sample 
* component and send transfered data to quantization component
* Author: Shengpu Liu
**************************************************************/

#include <sys_module.h>

#include <led_dbg.h>
#include "lswt.h"

#include <MassWare/MassWare/massware.h>

// no timer and blink display for this component.
#define BLINK_TIMER_INTERVAL	512L
#define BLINK_RED_TID         0

#define LSWT_SCALE_LEVEL  2
#define sqrt2             1.41421
#define sqrt3             1.73205

typedef struct {
  uint8_t pid;
  uint8_t state;
  MassOutputInterf * outfs;
  uint8_t outf_cnt;
  uint8_t lswt_scale_level;
  uint16_t data_length;
  data_type *data_set;
  uint16_t pkt_size; // the packet size and parcket are used to record the processed data and sent them
  void * pkt;        // after they are process and green LED blinked
} app_state_t;

static int8_t lswt_msg_handler(void *start, Message *e);
static void daub_forward_transfer(data_type *vec, uint16_t N, uint8_t nlevel);
/*
// static void daub_inverse_transfer(data_type *vec, uint16_t N, uint8_t nlevel);
static void split(data_type *vec, uint16_t N);
// static void merge(data_type *vec, uint16_t N);
static void normalize(data_type *S, uint16_t N, int direction);
static void updateOne(data_type *S, uint16_t N, int direction);
static void update(data_type *S, uint16_t N, int direction );
static void predict(data_type *S, uint16_t N, int direction );
static data_type round_t(double val);
*/
static const mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = LSW_TRANSFORM_MOD_ID,
	.state_size     = sizeof(app_state_t),
	.num_sub_func   = 0,
	.num_prov_func  = 0,
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id        = ehtons(LSW_TRANSFORM_MOD_ID),
	.module_handler = lswt_msg_handler,
};


static int8_t lswt_msg_handler(void *state, Message *msg)
{
  app_state_t *s = (app_state_t*)state;

  switch (msg->type){
  case MSG_INIT:
	  {
		  s->pid = msg->did;
		  s->state = 0;
		  s->outfs = NULL;
		  s->outf_cnt = 0;
		  s->lswt_scale_level = LSWT_SCALE_LEVEL;
		  s->data_length = 0;
		  s->data_set = NULL;

		  sys_led(LED_RED_OFF);
		  break;
	  }
  case MSG_SET_SCALE:
	  {
		  s->lswt_scale_level = (uint8_t)ehtonl(*((uint32_t*)(msg->data)));
		  break;
	  }
  case MSG_FORWARD_INPUT:
	  {
		  sys_led(LED_RED_OFF);

		  uint16_t i;
		  // input are uint16_t type from sensor module, we convert it to data_type
		  uint16_t * input = (uint16_t *)(msg->data);

		  // the first uint16 (2 bytes) is the data length, should be 2^n;
		  s->data_length = ehtons(input[0]);
		  // make the length to be 2^n;
		  uint8_t ln = 0;
		  while (s->data_length >= 2) {
			  s->data_length >>= 1;
			  ln++;
		  }
		  s->data_length <<= ln;

		  // get the packet size that should be sent out
		  s->pkt_size = sizeof(uint16_t) + + sizeof(uint8_t) + sizeof(data_type) * (s->data_length);
		  s->pkt = sys_malloc(s->pkt_size); 
		  if (NULL == s->pkt) break;

		  // the first two bytes are data length, the thrid byte is the scale number
		  // the rest are the data payload,
		  *(uint16_t*)(s->pkt) = s->data_length;
		  *((uint8_t*)(s->pkt) + 2) = s->lswt_scale_level;
		  s->data_set = (data_type *)((uint8_t *)s->pkt + 3);
		  for (i = 0; i<s->data_length; i++)
			  s->data_set[i] = ehtons(input[i+1]);

		  // transfer the data, transfered data are still in the original data set
		  daub_forward_transfer(s->data_set, s->data_length, s->lswt_scale_level);

		  DEBUG("Red Blink Start\n");
		  sys_led(LED_RED_TOGGLE);
		  sys_timer_start(BLINK_RED_TID, BLINK_TIMER_INTERVAL, TIMER_ONE_SHOT);
		  break;
	  }
  case MSG_FORWARD_OUTPUT:
	  {
		  // PostDestMsg *pdmsg = (PostDestMsg*)sys_msg_take_data(msg);
		  PostDestMsg *pdmsg = (PostDestMsg*)(msg->data);
		  MassOutputInterf * outf = search_interface(s->outfs, MSG_FORWARD_OUTPUT);
		  if (outf == NULL) {
			  outf = (MassOutputInterf*) sys_malloc(sizeof(MassOutputInterf));
			  outf->out_msg_type = MSG_FORWARD_OUTPUT;
			  outf->dst_mod_id = pdmsg->dst_mod_id;
			  outf->dst_msg_type = pdmsg->msg_type;
			  outf->next = s->outfs;
			  s->outfs = outf;
			  s->outf_cnt ++;
		  }
		  else {
			  outf->dst_mod_id = pdmsg->dst_mod_id;
			  outf->dst_msg_type = pdmsg->msg_type;
		  }
		  break;
	  }
	  /**
	  * All timers addressed to this PID, regardless of the timer ID, are of
	  * type MSG_TIMER_TIMEOUT and handled by this handler.  Timers with
	  * different timer IDs can be further distinguished by testing for the
	  * type, as demonstrated in the relay module.
	  */
  case MSG_TIMER_TIMEOUT:
	  {
		  sys_led(LED_RED_TOGGLE);
		  
		  // send the data set out, we do not need to release the data set because
		  // the send command will release the dataset
		  
		  MassOutputInterf * outf = search_interface(s->outfs, MSG_FORWARD_OUTPUT);
		  if (outf != NULL)
			  sys_post(outf->dst_mod_id, outf->dst_msg_type, s->pkt_size, s->pkt, SOS_MSG_RELEASE);
		  return SOS_OK;
	  }

  case MSG_INVERSE_INPUT:
	  break;
  case MSG_INVERSE_OUTPUT:
	  break;
  case MSG_FINAL:
	  {
		  // release memory
		  release_interface(s->outfs);
		  s->outfs = NULL;
		  s->outf_cnt = 0;
		  //sys_free(s->data_set); //no need free because the processed data are sent for release
		  s->data_length = 0;
		  s->data_set = NULL;
		  
		  // no led control in this component
		  break;
	  }
  default:
	  return -EINVAL;
  }

  return SOS_OK;
}

static void daub_forward_transfer(data_type *vec, uint16_t N, uint8_t nlevel)
{
/*	uint8_t i;
	//for (int n = N; n > 1; n = n >> 1) {
    for(i=0; i<nlevel; i++){
		uint16_t n = N>>i; //only trans nlevel steps
		split( vec, n );
		updateOne( vec, n, forward );  // update 1
		predict( vec, n, forward );
		update( vec, n, forward );     // update 2
		normalize( vec, n, forward );
	}
*/
} // forwardTrans, s->data_length, s->lswt_scale_level)

/****************************************************************
 * Default two step Lifting Scheme inverse wavelet transform
 * inverseTrans is passed the result of an ordered wavelet
 * transform, consisting of an average and a set of wavelet
 * coefficients.  The inverse transform is calculated
 * in-place and the result is returned in the argument array.
 ***************************************************************
static void daub_inverse_transfer(data_type *vec, uint16_t N, uint8_t nlevel)
{
    uint8_t i;
	//for (int n = 2; n <= N; n = n << 1) {
	for(i=1; i<=nlevel; i++){
		uint16_t n = N>>(nlevel-i); //only invtrans nlevel steps
		normalize( vec, n, inverse );
		update( vec, n, inverse );
		predict( vec, n, inverse );
		updateOne(vec, n, inverse );
		merge( vec, n );
	}
} // inverseTrans  
 ****************************************************************/


/*******************************************************
* Split the vec into even and odd elements,
* where the even elements are in the first half 
* of the vector and the odd elements are in the
* second half.
static void split(data_type *vec, uint16_t N)
{
	uint16_t start = 1;
    uint16_t end = N - 1;
	uint16_t i;

    while (start < end) {
		for (i = start; i < end; i = i + 2) {
			data_type tmp = vec[i];
			vec[i] = vec[i+1];
			vec[i+1] = tmp;
		}
		start = start + 1;
		end = end - 1;
	}
}
********************************************************/
  
/*******************************************************
 * Merge the odd elements from the second half of the N
 * element region in the array with the even elements in
 * the first half of the N element region.  The result
 * will be the combination of the odd and even elements
 * in a region of length N.
 *******************************************************
static void merge(data_type *vec, uint16_t N)
{
	uint16_t half = N >> 1;
    uint16_t start = half-1;
    uint16_t end = half;
	uint16_t i;

    while (start > 0) {
		for (i = start; i < end; i = i + 2)
		{
			data_type tmp = vec[i];
			vec[i] = vec[i+1];
			vec[i+1] = tmp;
		}
		start = start - 1;
		end = end + 1;
	}
}
********************************************************/
/*
static void normalize(data_type *S, uint16_t N, int direction)
{
	uint16_t half = N >> 1;
	uint16_t n;

    for (n = 0; n < half; n++) {
		if (direction == forward) {
			S[n] = (data_type)round_t(((sqrt3-1.0)/sqrt2) * S[n]);
			S[n+half] = (data_type)round_t(((sqrt3+1.0)/sqrt2) * S[n+half]);
		}
		else if (direction == inverse) {
			S[n] = (data_type)round_t(((sqrt3+1.0)/sqrt2) * S[n]);
			S[n+half] = (data_type)round_t(((sqrt3-1.0)/sqrt2) * S[n+half]);
		}
		else {
			return;
			// cerr<<"daubechies::normalize: bad direction value"<<endl;
			// exit(0);
		}
	}
} // normalize
  
static void updateOne(data_type *S, uint16_t N, int direction)
{
	uint16_t half = N >> 1;
	uint16_t n;

    for (n = 0; n < half; n++) {
		data_type updateVal = (data_type)round_t(sqrt3 * S[half+n]);
		if (direction == forward)
			S[n] = S[n] + updateVal;
		else if (direction == inverse)
			S[n] = S[n] - updateVal;
		else {
			// cerr<<"daubechies::updateOne: bad direction value"<<endl;
			// exit(0);
			return;
		}
	}
} // updateOne

static void update(data_type *S, uint16_t N, int direction )
{
	uint16_t half = N >> 1;
	uint16_t n;

    for (n = 0; n < half-1; n++) {
		if (direction == forward)
			S[n] = S[n] - S[half+n+1];
		else if (direction == inverse)
			S[n] = S[n] + S[half+n+1];
		else {
			// cerr<<"daubechies::update: bad direction value"<<endl;
			// exit(0);
			return;
		}
	}

    if (direction == forward)
		S[half-1] = S[half-1] - S[half];
	else if (direction == inverse)
		S[half-1] = S[half-1] + S[half];
} // update

static void predict(data_type *S, uint16_t N, int direction )
{
	uint16_t half = N >> 1;
	uint16_t n;

    if (direction == forward)
		S[half] = (data_type)round_t(S[half] - (sqrt3/4.0)*S[0] - (((sqrt3-2)/4.0)*S[half-1]));
	else if (direction == inverse)
		S[half] = (data_type)round_t(S[half] + (sqrt3/4.0)*S[0] +	(((sqrt3-2)/4.0)*S[half-1]));
	else return;
		// cerr<<"daubechies::predict: bad direction value"<<endl;

	// predict, forward
    for (n = 1; n < half; n++)
		if (direction == forward)
			S[half+n] = (data_type)round_t(S[half+n] - (sqrt3/4.0)*S[n] - (((sqrt3-2)/4.0)*S[n-1]));
		else if (direction == inverse)
			S[half+n] = (data_type)round_t(S[half+n] + (sqrt3/4.0)*S[n] + (((sqrt3-2)/4.0)*S[n-1]));
		else return;
}  // predict

// contained in math.h
static data_type round_t(double val)
{
	data_type ck = (data_type)val;
	if (ck == val) return ck;

    if(val >= 0)
		return (data_type)(val + 0.5);
	return (data_type)(val - 0.5);
}
*/

#ifndef _MODULE_
mod_header_ptr lswt_get_header()
{
  return sos_get_header_address(mod_header);
}
#endif


