/* -*- Mode: C; tab-width:4 -*- */
/* ex: set ts=4 shiftwidth=4 softtabstop=4 cindent: */
/**
 * @brief Example module for SOS
 * This module shows some of concepts in SOS
 */

/**
 * Module needs to include <module.h>
 */
//#undef _MODULE_

#include <sys_module.h>

//#define LED_DEBUG
#include <led_dbg.h>

#include "quantz.h"

#include <MassWare/MassWare/massware.h>

#define BLINK_TIMER_INTERVAL	512L
#define BLINK_GREEN_TID         0
#define QUAN_NQR     3     //for quantization 
#define QUAN_NUM     5000  //for quantization 
#define log2         0.693147

/**
 * Module can define its own state
 */
typedef struct {
  uint8_t pid;
  uint8_t state;
  MassOutputInterf * outfs;
  uint8_t outf_cnt;
  uint8_t lswt_scale_level; // scale level is used to calcuate the quantz value for each part
  uint16_t data_length;
  data_type *data_set;
  double *quan;      // quantization factors, we use the same one if the data length and lswt
  bool bInitialized; // check if the quantization factors are calculated and fresh
  uint16_t pkt_size; // the packet size and parcket are used to record the processed data and sent them
  void * pkt;        // after they are process and green LED blinked
} app_state_t;

static int8_t quantization_msg_handler(void *start, Message *e);
static double * init_quantization_factor(data_type *vec, uint16_t N, uint8_t nlevel);
/*
static double * getMaxMinHist(data_type *ary, uint16_t ln, data_type *pmax, data_type *pmin, data_type *pdif);
static uint8_t quanIndex(uint16_t index, uint16_t len, uint8_t nlevel);
static data_type round_t(double val);
//static int pow2(uint8_t M);
static double pow_int(double X, int Y);
static double factorial(uint8_t N);
static double log(double val);
static double exp(double X);
static double pow(double base, double idx);
*/

/**
 * This is the only global variable one can have.
 */
static const mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = SCALE_QUANTIZATION_MOD_ID,
	.state_size     = sizeof(app_state_t),
	.num_sub_func   = 0,
	.num_prov_func  = 0,
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id        = ehtons(SCALE_QUANTIZATION_MOD_ID),
	.module_handler = quantization_msg_handler,
};


static int8_t quantization_msg_handler(void *state, Message *msg)
{
	/**
	* The module is passed in a void* that contains its state.  For easy
	* reference it is handy to typecast this variable to be of the
	* applications state type.  Note that since we are running as a module,
	* this state is not accessible in the form of a global or static
	* variable.
	*/
	app_state_t *s = (app_state_t*)state;

	/**
	* Switch to the correct message handler
	*/
	switch (msg->type){
		/**
		* MSG_INIT is used to initialize module state the first time the
		* module is used.  Many modules set timers at this point, so that
		* they will continue to receive periodic (or one shot) timer events.
		*/
  case MSG_INIT:
	  {
		  s->pid = msg->did;
		  s->state = 0;
		  s->outfs = NULL;
		  s->outf_cnt = 0;
		  s->lswt_scale_level = 0;
		  s->data_length = 0;
		  s->data_set = NULL;
		  s->quan = NULL;
		  s->bInitialized = false;

		  sys_led(LED_GREEN_OFF);
		  /**
		  * The timer API takes the following parameters:
		  * - Timer ID (used to distinguish multiple timers of different
		  *   ..types on the same host)
		  * - Timer delay in
		  * - Type of timer
		  */
		  //	  DEBUG("Green Blink Start\n");
		  //	  sys_timer_start(BLINK_GREEN_TID, BLINK_TIMER_INTERVAL, TIMER_REPEAT);
		  break;
	  }

  case MSG_FORWARD_INPUT:
	  {
		  sys_led(LED_GREEN_OFF);

		  uint16_t i;

		  // input are data_type type from lswt module, we keep it as data_type

		  // the first uint16 (2 bytes) is the data length, the thrid byte is the scale number
		  // the rest are the data payload,
		  uint16_t len = ehtons(*(uint16_t *)(msg->data));
		  if (s->data_length != len) {
			  if (s->bInitialized) {
				  sys_free(s->quan);
				  s->bInitialized = false;
			  }
			  s->data_length = len;
		  }

		  uint8_t scale = *((uint8_t *)(msg->data) + 2);
		  if (scale != s->lswt_scale_level) {
			  if (s->bInitialized) {
				  sys_free(s->quan);
				  s->bInitialized = false;
			  }
			  s->lswt_scale_level = scale;
		  }

		  data_type * input = (data_type *)((uint8_t *)(msg->data) + 3);

		  // get the packet size that should be sent out
		  s->pkt_size = sizeof(uint16_t) + sizeof(uint8_t) + sizeof(data_type) * (s->data_length);
		  s->pkt = sys_malloc(s->pkt_size); 
		  if (NULL == s->pkt) break;

		  // the first two bytes are data length, the thrid byte is the scale number
		  // the rest are the data payload,
		  *(uint16_t*)(s->pkt) = ehtons(s->data_length);
		  *((uint8_t*)(s->pkt) + 2) = s->lswt_scale_level;
		  s->data_set = (data_type *)((uint8_t *)(s->pkt) + 3);
		  for (i = 0; i<s->data_length; i++)
			  s->data_set[i] = ehtons(input[i]);

		  if (!s->bInitialized) {
			  s->quan = init_quantization_factor(s->data_set, s->data_length, s->lswt_scale_level);
			  s->bInitialized = true;
		  }

		  // quantize the data, quantized data are still in the original data set
		  for (i = 0; i<s->data_length; i++)
			  s->data_set[i] = 0; //round_t( s->data_set[i] / s->quan[quanIndex(i, s->data_length, s->lswt_scale_level)] );
		  

		  DEBUG("Green Blink Start\n");
		  sys_led(LED_GREEN_TOGGLE);
		  sys_timer_start(BLINK_GREEN_TID, BLINK_TIMER_INTERVAL, TIMER_ONE_SHOT);
		  break;
	  }
	  
	  /**
	  * set the output component id (mod_id) and the interface of the component (msg_type)
	  * so that masslets can be connected dynamically, used for arch reconfiguration
	  */    

  case MSG_FORWARD_OUTPUT:
	  {
		  // PostDestMsg *pdmsg = (PostDestMsg*)sys_msg_take_data(msg);
		  PostDestMsg *pdmsg = (PostDestMsg*)(msg->data);
		  MassOutputInterf * outf = search_interface(s->outfs, MSG_FORWARD_OUTPUT);
		  if (outf == NULL) {
			  outf = (MassOutputInterf*) sys_malloc(sizeof(MassOutputInterf));
			  outf->out_msg_type = MSG_FORWARD_OUTPUT;
			  outf->dst_mod_id = pdmsg->dst_mod_id;
			  outf->dst_msg_type = pdmsg->msg_type;
			  outf->next = s->outfs;
			  s->outfs = outf;
			  s->outf_cnt ++;
		  }
		  else {
			  outf->dst_mod_id = pdmsg->dst_mod_id;
			  outf->dst_msg_type = pdmsg->msg_type;
		  }
		  break;
	  }
  case MSG_INVERSE_INPUT:
	  break;
  case MSG_INVERSE_OUTPUT:
	  break;
	  
	  /**
	  * All timers addressed to this PID, regardless of the timer ID, are of
	  * type MSG_TIMER_TIMEOUT and handled by this handler.  Timers with
	  * different timer IDs can be further distinguished by testing for the
	  * type, as demonstrated in the relay module.
	  */
  case MSG_TIMER_TIMEOUT:
	  {
		  sys_led(LED_GREEN_TOGGLE);
		  
		  // send the data set out, we do not need to release the data set because
		  // the send command will release the dataset
		  
		  MassOutputInterf * outf = search_interface(s->outfs, MSG_FORWARD_OUTPUT);
		  if (outf != NULL)
			  sys_post(outf->dst_mod_id, outf->dst_msg_type, s->pkt_size, s->pkt, SOS_MSG_RELEASE);
		  return SOS_OK;
	  }
	  
	  /**
	  * MSG_FINAL is used to shut modules down.  Modules should release all
	  * resources at this time and take care of any final protocol
	  * shutdown.
	  */
  case MSG_FINAL:
	  {
		  // release memory
		  release_interface(s->outfs);
		  s->outfs = NULL;
		  s->outf_cnt = 0;

		  // free quantization factors
		  sys_free(s->quan);

		  sys_led(LED_GREEN_OFF);
		  DEBUG("Green Blink Stop\n");
		  break;
	  }

	  /**
	  * The default handler is used to catch any messages that the module
	  * does no know how to handle.
	  */
  default:
	  return -EINVAL;
  }

  /**
  * Return SOS_OK for those handlers that have successfully been handled.
  */
  return SOS_OK;
}

static double * init_quantization_factor(data_type *vec, uint16_t N, uint8_t nlevel)
{
	double * quan = (double *) sys_malloc(sizeof(double) * (nlevel + 1));
/*	double * var  = (double *) sys_malloc(sizeof(double) * (nlevel + 1));
	double * hist;
	data_type amax, amin, ndif;

	uint8_t j;
	uint16_t k, len = N>>nlevel;

    for(j=0; j<=nlevel; j++){
		if (j==0)
			hist = getMaxMinHist(vec, len, &amax, &amin, &ndif);
		else
			hist = getMaxMinHist(vec+len, len, &amax, &amin, &ndif);
		for(k=0; k<ndif; k++)
			hist[k] = hist[k]/len;
		var[j] = 0;
		for (k=0; k<ndif; k++)
			var[j] = var[j] + (amin+k)*(amin+k)*hist[k];
		sys_free(hist);
		if(j>0) len<<=1;

	}

	double fac = 1.0;
	for (j=0; j<=nlevel; j++)
		fac = fac * (double)pow((double)var[j], 0.25);

	for (j=0; j<=nlevel; j++)
		quan[j] = QUAN_NQR + log(var[j]/fac)/log2/2;

	sys_free(var);

	for (j=0; j<=nlevel; j++)
		quan[j]=QUAN_NUM/pow(2, quan[j]);
*/
	return quan;
}
/*
static double * getMaxMinHist(data_type *ary, uint16_t ln, data_type *pmax, data_type *pmin, data_type *pdif)
{
	uint16_t i;
	double *hst;
	*pmax = ary[0];
	*pmin = ary[0];

	for(i=1; i<ln; i++) {
		if(ary[i] > *pmax) 
			*pmax = ary[i];
		else if(ary[i] < *pmin) 
			*pmin = ary[i];
	}

	*pdif = *pmax - *pmin + 1;
	
	hst = (double *) sys_malloc((*pdif) * sizeof(double));
	for (i=0; i< *pdif; i++)
		hst[i] = 0;
	for(i=0; i<ln; i++)
		hst[ary[i]-(*pmin)] += 1.0;

	return hst;
}

static uint8_t quanIndex(uint16_t index, uint16_t len, uint8_t nlevel)
{
	uint8_t i;
	for(i=0; i<=nlevel; i++) {
		if( index < (len>>(nlevel-i)) ) 
			return i;
	}
	return -1;
}

// contained in math.h
static data_type round_t(double val)
{
	data_type ck = (data_type)val;
	if (ck == val) return ck;

    if(val >= 0)
		return (data_type)(val + 0.5);
	return (data_type)(val - 0.5);
}
*/
// 2^M
/*
static int pow2(uint8_t M)
{
	int res = 1;
	uint8_t i;
	for (i = 0; i<M; i++)
		res = res * 2;
	return res;
}

// X^Y
static double pow_int(double X, int Y)
{
	uint8_t i;
	double res = 1.0;
	for (i = 0; i<Y; i++)
		res = res * X;
	return res;
}

// N!
static double factorial(uint8_t N)
{
	uint8_t i;
	double res = 1.0;
	for (i = 2; i <= N; i++)
		res = res * i;
	return res;
}

//http://en.wikipedia.org/wiki/Natural_logarithm
double log(double X)
{
	double res = 0.0;
	uint8_t i;
	if (0 == X) 
		return res; // this is an error
	else if (X > 0 && X <= 2) {
		for (i=1; i<=P; i++) {
			if (i & 1)
				res = res + pow_int(X-1, i)/i;
			else 
				res = res - pow_int(X-1, i)/i;
		}
	}
	else {
		res = 0 - log(1/(X-1));
		for (i=1; i<=P; i++) {
			res = res + 1.0/(i * pow_int(X, i));
		}
	}
	return res;
}
*/

/****************************************************
e^x = 1+x+ (x^2/2!) + (x^3/3!) +.......+ (x^n/n!)
static double exp(double X)
{
	uint8_t i;
	double result = 1 + X;
	for (i = 2; i<= P; i++) {
		result += (pow_int(X, i) / factorial(i));
	}
	return result;
}

//X^Y = exp(Y * log(X));
static double pow(double X, double Y)
{
	return exp(Y * log(X));
}
*****************************************************/


#ifndef _MODULE_
mod_header_ptr quantization_get_header()
{
	return sos_get_header_address(mod_header);
}
#endif


