/**************************************************************
* lswt.c
* This file implements the lifting scheme wavelet transform 
* algorithm for wireless sensor network based on SOS
* It is designed as a component that gets input from a sample 
* component and send transfered data to quantization component
* Author: Shengpu Liu
**************************************************************/

#include <sys_module.h>

#include <led_dbg.h>
#include "sensing.h"

#include <MassWare/MassWare/massware.h>
#include <sensordrivers/mts310/include/mts310sb.h>

// no timer and blink display for this component.
#define SAMPLE_DATA_LENGTH  4
#define SAMPLE_FREQENCY     256L
#define SAMPLE_REPEAT_TID   0

typedef struct {
  uint8_t pid;
  uint8_t state;
  MassOutputInterf * outfs;
  uint8_t outf_cnt;
  uint16_t data_length;
  data_type *data_set;
  int32_t sample_freq;
  uint16_t counter;
  bool bStarted; 
} app_state_t;

static int8_t sensing_msg_handler(void *start, Message *e);

static const mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = SENSOR_READING_MOD_ID,
	.state_size     = sizeof(app_state_t),
	.num_sub_func   = 0,
	.num_prov_func  = 0,
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id        = ehtons(SENSOR_READING_MOD_ID),
	.module_handler = sensing_msg_handler,
};


static int8_t sensing_msg_handler(void *state, Message *msg)
{
  app_state_t *s = (app_state_t*)state;

  switch (msg->type){
  case MSG_INIT:
	  {
		  s->pid = msg->did;
		  s->state = 0;
		  s->outfs = NULL;
		  s->outf_cnt = 0;
		  s->data_length = SAMPLE_DATA_LENGTH;
		  s->data_set = sys_malloc(sizeof(data_type) * s->data_length);
		  s->counter = 0;
		  s->sample_freq = SAMPLE_FREQENCY;
		  s->bStarted = false;

		  sys_led(LED_RED_OFF);
		  break;
	  }
  case MSG_SET_DATA_LENGTH:
	  {
		  uint16_t len = (uint16_t)ehtonl(*((uint32_t*)(msg->data)));
		  if (len != s->data_length) {
			  s->data_length = len;
			  if (NULL != s->data_set) {
				  sys_free(s->data_set);
				  s->data_set = sys_malloc(sizeof(data_type) * len);
				  s->counter = 0;
			  }
		  }
		  break;
	  }
  case MSG_SET_SAMPEL_FREQ:
	  {
		  s->sample_freq = (int32_t)ehtonl(*((uint32_t*)(msg->data)));
		  if (s->bStarted) 
			  sys_timer_restart(SAMPLE_REPEAT_TID, s->sample_freq);
		  break;
	  }
  case MSG_START_SAMPLE:
	  {
		  sys_led(LED_RED_OFF);
		  sys_timer_start(SAMPLE_REPEAT_TID, s->sample_freq, TIMER_REPEAT);
		  s->bStarted = true;
		  break;
	  }

  case MSG_TIMER_TIMEOUT:
	  {
		  if(sys_sensor_get_data(MTS310_PHOTO_SID) < 0)
 {
    		  sys_led(LED_RED_TOGGLE);
			  sys_post_value(SENSOR_READING_MOD_ID, MSG_DATA_READY, 0x00ffff00, 0);
	      } 
		  else {
    		  sys_led(LED_GREEN_TOGGLE);
		  }

		  return SOS_OK;
	  }

  case MSG_DATA_READY:
	  {
		  MsgParam* param = (MsgParam*) (msg->data);

		  s->data_set[s->counter++] = ehtons(param->word);

		  if (s->counter == s->data_length) {
			  // the first two bytes are the data length, the rest are the payload
			  uint16_t pkt_size = sizeof (uint16_t) + s->data_length * sizeof(data_type);
			  void * pkt = sys_malloc(pkt_size);

			  *(uint16_t*)pkt = ehtons(s->data_length);
			  data_type * output = (data_type *)((uint16_t*)pkt + 1);
			  
			  uint16_t i;
			  for (i = 0; i < s->data_length; i++) 
				  output[i] = ehtons(s->data_set[i]);

			  MassOutputInterf * outf = search_interface(s->outfs, MSG_OUTPUT_DATA);
			  if (outf != NULL)
				  sys_post(outf->dst_mod_id, outf->dst_msg_type, pkt_size, pkt, SOS_MSG_RELEASE);

			  s->counter = 0;
		  }
	  }
		  
  case MSG_OUTPUT_DATA:
	  {
		  // PostDestMsg *pdmsg = (PostDestMsg*)sys_msg_take_data(msg);
		  PostDestMsg *pdmsg = (PostDestMsg*)(msg->data);
		  MassOutputInterf * outf = search_interface(s->outfs, MSG_OUTPUT_DATA);
		  if (outf == NULL) {
			  outf = (MassOutputInterf*) sys_malloc(sizeof(MassOutputInterf));
			  outf->out_msg_type = MSG_OUTPUT_DATA;
			  outf->dst_mod_id = pdmsg->dst_mod_id;
			  outf->dst_msg_type = pdmsg->msg_type;
			  outf->next = s->outfs;
			  s->outfs = outf;
			  s->outf_cnt ++;
		  }
		  else {
			  outf->dst_mod_id = pdmsg->dst_mod_id;
			  outf->dst_msg_type = pdmsg->msg_type;
		  }
		  break;
	  }
  case MSG_FINAL:
	  {
		  // release memory
		  release_interface(s->outfs);
		  s->outfs = NULL;
		  s->outf_cnt = 0;
		  
		  sys_free(s->data_set); //no need free because the processed data are sent for release
		  s->data_length = 0;
		  s->data_set = NULL;
		  
		  // no led control in this component
		  break;
	  }
  default:
	  return -EINVAL;
  }

  return SOS_OK;
}


#ifndef _MODULE_
mod_header_ptr sensing_get_header()
{
  return sos_get_header_address(mod_header);
}
#endif


